from copy import deepcopy

from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    process_classes,
    GenericDirective,
    generic_visit,
    generic_depart,
    get_node,
    get_generic_directive,
    BaseClassNode,
    BaseNode,
)

class pseudo_heading_node(BaseClassNode): pass
class text_landing_node(BaseClassNode):
    classes=['text-landing']

def setup_heading(app):
    app.add_node(pseudo_heading_node)
    app.add_node(text_landing_node)
    app.add_role('ph1', heading_role(pseudo_heading_node, 'h1'))
    app.add_role('ph2', heading_role(pseudo_heading_node, 'h2'))
    app.add_role('ph3', heading_role(pseudo_heading_node, 'h3'))
    app.add_role('ph4', heading_role(pseudo_heading_node, 'h4'))
    app.add_role('ph5', heading_role(pseudo_heading_node, 'h5'))
    app.add_role('ph6', heading_role(pseudo_heading_node, 'h6'))
    app.add_role('tlh1', heading_role(text_landing_node, 'h1'))
    app.add_role('tlh2', heading_role(text_landing_node, 'h2'))
    app.add_role('tlh3', heading_role(text_landing_node, 'h3'))
    app.add_role('tlh4', heading_role(text_landing_node, 'h4'))
    app.add_role('tlh5', heading_role(text_landing_node, 'h5'))
    app.add_role('tlh6', heading_role(text_landing_node, 'h6'))
    app.add_directive('heading', HeadingDirective)
    app.add_directive('text-landing', TextLandingDirective)


def heading_role(node_type, tag: str):
    def role(name, rawtext, text, lineno, inliner, options={}, content=[]):
        node = node_type()
        node += nodes.Text(text)
        node['tag'] = tag
        return [node], []
    return role

class HeadingDirective(GenericDirective):
    required_arguments = 1
    node_type = pseudo_heading_node

    def run(self):
        node_list = super().run()
        node_list[0]['tag'] = self.arguments[0]
        return node_list

class TextLandingDirective(HeadingDirective):
    node_type = text_landing_node
    option_spec = deepcopy(HeadingDirective.option_spec)
    option_spec.update({
        'shadow': directives.unchanged,
    })

    def run(self):
        node = super().run()[0]
        if 'shadow' in self.options:
            node.add_class('text-shadow')
        return [node]

class HeadingTranslator(HTMLTranslator):

    def visit_pseudo_heading_node(self, node):
        generic_visit(self, node['tag'], node)
    
    def depart_pseudo_heading_node(self, node):
        generic_depart(self, node['tag'])
    
    def visit_text_landing_node(self, node):
        generic_visit(self, node['tag'], node)
    
    def depart_text_landing_node(self, node):
        generic_depart(self, node['tag'])


from copy import deepcopy

from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    BaseClassNode,
    add_class_to_tag,
    GenericDirective,
)

class fa_node(BaseClassNode):
    pass

class fa_large_node(BaseClassNode):
    classes = ['large']

class fa_span_node(nodes.Element):
    pass

def fa_role_maker(span: bool=False, large: bool = False):
    def fa_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
        if span:
            # TODO: this should really add a span node with fa_node inside
            node = fa_span_node()
        elif large:
            node = fa_large_node()
        else:
            node = fa_node()
        node['fa_type'] = text
        return [node], []
    return fa_role

class FontAwesomeDirective(GenericDirective):
    required_arguments = 1
    optional_arguments = 4

    option_spec = deepcopy(GenericDirective.option_spec)
    option_spec.update({
        'span': directives.unchanged,
        'large': directives.unchanged,
    })

    def run(self):
        # TODO: This should be handled better, i.e.
        # span node containing fa node
        if 'span' in self.options:
            self.node_type = fa_span_node
        elif 'large' in self.options:
            self.node_type = fa_large_node
        else:
            self.node_type = fa_node
        node = super().run()[0]
        node['fa_type'] = self.arguments[0]
        return [node]

def setup_fa(app):
    app.add_node(fa_node)
    app.add_node(fa_span_node)
    app.add_role('font-awesome', fa_role_maker())
    app.add_role('font-awesome-span', fa_role_maker(span=True))
    app.add_directive('font-awesome', FontAwesomeDirective)
    # Convenience
    app.add_role('fa', fa_role_maker())
    app.add_role('fa-span', fa_role_maker(span=True))
    app.add_role('fa-l', fa_role_maker(large=True))

class FontAwesomeTranslator(HTMLTranslator):
    
    def visit_fa_node(self, node):
        fa_type = node['fa_type']
        content = f"""<svg class=\"{fa_type}\">
    <use xlink:href=\"#{fa_type}\"></use>
</svg>"""
        for class_name in node.get('classes', []):
            content = add_class_to_tag(content, 'svg', class_name)
        self.body.append(content)

    def visit_fa_span_node(self, node):
        self.body.append('<span>')
        self.visit_fa_node(node)

    def depart_fa_node(self, node):
        pass
    
    def depart_fa_span_node(self, node):
        self.body.append('</span>')

    def visit_fa_large_node(self, node):
        self.visit_fa_node(node)

    def depart_fa_large_node(self, node):
        self.depart_fa_node(node)


from sphinx.writers.html import HTMLWriter

from .slide import SlideTranslator
from .font_awesome import FontAwesomeTranslator
from .heading import HeadingTranslator
from .header_footer import HeaderFooterTranslator
from .div import DivTranslator
from .span import SpanTranslator
from .paragraph import ParagraphTranslator
from .link import LinkTranslator
from .flexblock import FlexblockTranslator
from .hr import HRTranslator
from .description_list import DescriptionListTranslator
from .preformatted import PreformattedTranslator
from .figure import FigureTranslator
from .web import WebTranslator
from .video import VideoTranslator
from .blockquote import BlockquoteTranslator
from .template import TemplateTranslator

class WebslidesTranslator(
    SlideTranslator,
    HeadingTranslator,
    FontAwesomeTranslator,
    DivTranslator,
    SpanTranslator,
    ParagraphTranslator,
    LinkTranslator,
    FlexblockTranslator,
    HRTranslator,
    DescriptionListTranslator,
    PreformattedTranslator,
    HeaderFooterTranslator,
    FigureTranslator,
    WebTranslator,
    VideoTranslator,
    BlockquoteTranslator,
    TemplateTranslator,
):

    def visit_section(self, node):
        super().visit_section(node)
        # The div section that is added is problematic
        # Removing the div but still running to maintain
        # other state info
        self.body.pop()

    def depart_section(self, node):
        super().depart_section(node)
        # The div section that is added is problematic
        # Removing the div but still running to maintain
        # other state info
        self.body.pop()

    def visit_literal_block(self, node):
        self.body.append(f'<code>')

    def depart_literal_block(self, node):
        self.body.append('</code>')

class WebslidesWriter(HTMLWriter):
    translator_class = WebslidesTranslator


from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    add_class_to_tag,
    BaseClassNode,
    GenericDirective,
)

class description_list_node(BaseClassNode):
    classes = ['description']

class DescriptionListDirective(GenericDirective):
    node_type = description_list_node

def setup_description_list(app):
    app.add_node(description_list_node)
    app.add_directive('description-list', DescriptionListDirective)

class DescriptionListTranslator(HTMLTranslator):

    def visit_description_list_node(self, node):
        self.description_list_open = True
        content = '<ul>'
        for class_name in node['classes']:
            content = add_class_to_tag(
                content, 'ul', class_name
            )
        self.body.append(content)
    
    def depart_bullet_list(self, node):
        if not self.description_list_open:
            super().depart_bullet_list(node)

    def depart_description_list_node(self, node):
        self.description_list_open = False
        self.body.append('</ul>')


import copy

from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    add_class_to_tag,
    GenericDirective,
    BaseNode,
    BaseClassNode,
)

class flexblock_list_node(BaseNode): pass
class flexblock_list_border_node(BaseClassNode, flexblock_list_node):
    classes = ['border']
class flexblock_list_reasons_node(BaseClassNode, flexblock_list_node):
    classes = ['reasons']
class flexblock_list_steps_node(BaseClassNode, flexblock_list_node):
    classes = ['steps']
class flexblock_list_clients_node(BaseClassNode, flexblock_list_node):
    classes = ['clients']
class flexblock_list_features_node(BaseClassNode, flexblock_list_node):
    classes = ['features']
class flexblock_list_metrics_node(BaseClassNode, flexblock_list_node):
    classes = ['metrics']
class flexblock_list_activity_node(BaseClassNode, flexblock_list_node):
    classes = ['activity']
class flexblock_list_gallery_node(BaseClassNode, flexblock_list_node):
    classes = ['gallery']
class flexblock_list_spec_node(BaseClassNode, flexblock_list_node):
    classes = ['specs']
class text_cols_list_node(BaseNode): pass

class FlexblockDirective(GenericDirective):
    node_type = flexblock_list_node
class FlexblockBorderDirective(GenericDirective):
    node_type = flexblock_list_border_node
class FlexblockReasonsDirective(GenericDirective):
    node_type = flexblock_list_reasons_node
class FlexblockStepsDirective(GenericDirective):
    node_type = flexblock_list_steps_node
class FlexblockClientsDirective(GenericDirective):
    node_type = flexblock_list_clients_node
class FlexblockFeaturesDirective(GenericDirective):
    node_type = flexblock_list_features_node
class FlexblockMetricsDirective(GenericDirective):
    node_type = flexblock_list_metrics_node
class FlexblockActivityDirective(GenericDirective):
    node_type = flexblock_list_activity_node
class FlexblockGalleryDirective(GenericDirective):
    node_type = flexblock_list_gallery_node
class FlexblockSpecDirective(GenericDirective):
    node_type = flexblock_list_spec_node
class TextColsDirective(GenericDirective):
    node_type = text_cols_list_node

def setup_flexblock(app):
    app.add_node(flexblock_list_node)
    app.add_node(flexblock_list_border_node)
    app.add_node(flexblock_list_reasons_node)
    app.add_node(flexblock_list_steps_node)
    app.add_node(flexblock_list_clients_node)
    app.add_node(flexblock_list_features_node)
    app.add_node(flexblock_list_metrics_node)
    app.add_node(flexblock_list_activity_node)
    app.add_node(flexblock_list_gallery_node)
    app.add_node(flexblock_list_spec_node)
    app.add_node(text_cols_list_node)
    app.add_directive('flexblock', FlexblockDirective)
    app.add_directive('flexblock-border', FlexblockBorderDirective)
    app.add_directive('flexblock-reasons', FlexblockReasonsDirective)
    app.add_directive('flexblock-steps', FlexblockStepsDirective)
    app.add_directive('flexblock-clients', FlexblockClientsDirective)
    app.add_directive('flexblock-features', FlexblockFeaturesDirective)
    app.add_directive('flexblock-metrics', FlexblockMetricsDirective)
    app.add_directive('flexblock-activity', FlexblockActivityDirective)
    app.add_directive('flexblock-gallery', FlexblockGalleryDirective)
    app.add_directive('flexblock-spec', FlexblockSpecDirective)
    app.add_directive('text-cols', TextColsDirective)

class FlexblockTranslator(HTMLTranslator):
    flexblock_open = False
    flexblock_classes = []
    description_list_open = False
    text_cols_open = False
    flexblock_div_open = False

    def visit_flexblock_list_node(self, node):
        self.flexblock_div_open = True
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_node(self, node):
        self.base_depart_flexblock_list_node(node)
        self.flexblock_div_open = False
    
    def base_visit_flexblock_list_node(self, node):
        self.flexblock_open = True
        self.flexblock_classes = copy.deepcopy(node['classes'])

    def base_depart_flexblock_list_node(self, node):
        pass

    def visit_flexblock_list_border_node(self, node):
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_border_node(self, node):
        self.base_depart_flexblock_list_node(node)
    
    def visit_flexblock_list_reasons_node(self, node):
        self.body.append("<div class=\"bg-white shadow\">")
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_reasons_node(self, node):
        self.base_depart_flexblock_list_node(node)
        self.body.append("</div>")
    
    def visit_flexblock_list_steps_node(self, node):
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_steps_node(self, node):
        self.base_depart_flexblock_list_node(node)
        
    def visit_flexblock_list_clients_node(self, node):
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_clients_node(self, node):
        self.base_depart_flexblock_list_node(node)
    
    def visit_flexblock_list_features_node(self, node):
        self.flexblock_div_open = True
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_features_node(self, node):
        self.flexblock_div_open = False
        self.base_depart_flexblock_list_node(node)
    
    def visit_flexblock_list_metrics_node(self, node):
        self.flexblock_div_open = True
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_metrics_node(self, node):
        self.flexblock_div_open = False
        self.base_depart_flexblock_list_node(node)
    
    def visit_flexblock_list_activity_node(self, node):
        self.flexblock_div_open = True
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_activity_node(self, node):
        self.flexblock_div_open = False
        self.base_depart_flexblock_list_node(node)
    
    def visit_flexblock_list_gallery_node(self, node):
        self.flexblock_div_open = True
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_gallery_node(self, node):
        self.flexblock_div_open = False
        self.base_depart_flexblock_list_node(node)
    
    def visit_flexblock_list_spec_node(self, node):
        self.flexblock_div_open = True
        self.base_visit_flexblock_list_node(node)

    def depart_flexblock_list_spec_node(self, node):
        self.flexblock_div_open = False
        self.base_depart_flexblock_list_node(node)
    
    def visit_text_cols_list_node(self, node):
        self.text_cols_open = True

    def depart_text_cols_list_node(self, node):
        pass

    def visit_bullet_list(self, node):
        # TODO: not sure how to handle this better
        # Not portable as written
        # Could consider combining
        if not self.description_list_open:
            super().visit_bullet_list(node)
            if self.flexblock_open:
                self.body[-1] = add_class_to_tag(self.body[-1], 'ul', 'flexblock')
                for class_name in self.flexblock_classes:
                    self.body[-1] = add_class_to_tag(self.body[-1], 'ul', class_name)
                self.flexblock_open = False
                self.flexblock_classes = []
            elif self.text_cols_open:
                self.body[-1] = add_class_to_tag(self.body[-1], 'ul', 'text-cols')
                self.text_cols_open = False

    def visit_list_item(self, node):
        super().visit_list_item(node)
        if self.flexblock_div_open:
            self.body.append('<div>')

    def depart_list_item(self, node):
        if self.flexblock_div_open:
            self.body.append('</div>')
        super().depart_list_item(node)



from copy import deepcopy

from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    process_classes,
    GenericDirective,
    add_attribute_to_tag,
    create_role,
    get_generic_directive,
    add_role_and_directive,
    get_generic_translator_class,
    get_node,
    create_role,
    BaseNode,
    BaseClassNode,
    add_visit_depart,
)

TAG = 'span'

class span_node(BaseClassNode):pass
class text_label_node(BaseClassNode):
    classes = ['text-label']
class try_node(BaseClassNode):
    classes = ['try']
class span_raw_node(span_node): pass

ALLOWED_BACKGROUND_LOCATIONS = [
    'right-bottom',
    'left-bottom',
]
class background_node(span_node):
    classes=['background']

    def __init__(self, *args, 
        dark: bool = False, 
        light: bool = False, 
        animation: bool = False, 
        image_url: str = None, 
        location: str = None,
        **kwargs):
        super().__init__(*args, **kwargs)
        if dark:
            self.set_dark()
        if light:
            self.set_light()
        if animation:
            self.set_animation()
        if image_url is not None:
            self.set_background_image(image_url)
        if location is not None:
            self.add_location(location)

    def set_background_image(self, url: str):
        self['style'] = f"background-image:url('{url}')"

    def set_dark(self):
        self.add_class('dark')
    
    def set_light(self):
        self.add_class('light')

    def set_animation(self):
        self.add_class('anim')

    def add_location(self, requested_location):
        if requested_location in ALLOWED_BACKGROUND_LOCATIONS:
            self.add_class(f'background-{requested_location}')
            self['classes'].pop(self['classes'].index('background'))



class BackgroundImageDirective(GenericDirective):
    node_type = background_node
    required_arguments = 1
    option_spec = deepcopy(GenericDirective.option_spec)
    option_spec.update({
        'dark': directives.unchanged,
        'light': directives.unchanged,
        'animation': directives.unchanged,
        'location': directives.unchanged,
    })

    def run(self):
        node = super().run()[0]
        node.set_background_image(
            self.arguments[0]
        )
        if 'dark' in self.options:
            node.set_dark()
        if 'light' in self.options:
            node.set_light()
        if 'animation' in self.options:
            node.set_animation()
        loc = 'location'
        if loc in self.options:
            requested_location = self.options[loc].lower()
            node.add_location(requested_location)
        return [node]

def setup_span(app):
    app.add_node(span_node)
    app.add_node(span_raw_node)
    app.add_node(background_node)
    app.add_node(text_label_node)
    app.add_node(try_node)
    add_role_and_directive(app, try_node, 'try')
    add_role_and_directive(app, span_node, 'span')
    app.add_role('span-raw', raw_span_role)
    add_role_and_directive(app, text_label_node, 'text-label')
    app.add_directive('background-image', BackgroundImageDirective)
    app.add_role('background-image', create_role(background_node))
    
    # Convenience
    add_role_and_directive(app, text_label_node, 'tl')

def raw_span_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    node = span_raw_node()
    node['content'] = text
    return [node], []

class SpanTranslator(HTMLTranslator):
    
    def visit_span_node(self, node):
        content = f'<{TAG}>'
        content = process_classes(content, node, TAG)
        if 'style' in node:
            content = add_attribute_to_tag(
                content,
                TAG,
                'style',
                node['style']
            )
        self.body.append(content)
    
    def depart_span_node(self, node):
        self.body.append(f'</{TAG}>')

    def visit_span_raw_node(self, node):
        self.body.append(f"<{TAG}>{node['content']}")
    
    def depart_span_raw_node(self, node):
        self.depart_span_node(node)

    def visit_text_label_node(self, node):
        self.visit_span_node(node)

    def depart_text_label_node(self, node):
        self.depart_span_node(node)
        
    def visit_try_node(self, node):
        self.visit_span_node(node)

    def depart_try_node(self, node):
        self.depart_span_node(node)




import copy

from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    process_classes,
    GenericDirective,
    generic_visit,
    generic_depart,
    get_node,
    get_generic_directive,
    get_generic_translator_class,
    add_visit_depart,
    add_role_and_directive,
    BaseNode,
    BaseClassNode,
)

TAG = 'div'

class div_node(BaseClassNode): pass
class div_wrap_node(BaseClassNode):
    classes=['wrap']
class grid_node(BaseClassNode):
    classes=['grid']
class column_node(BaseClassNode):
    classes=['column']
class center_node(BaseClassNode):
    classes=['aligncenter']
class flex_content_node(BaseClassNode):
    classes=['flex-content']
class content_left_node(BaseClassNode):
    classes=['content-left']
class content_center_node(BaseClassNode):
    classes=['content-center']
class content_right_node(BaseClassNode):
    classes=['content-right']
class overlay_node(BaseClassNode):
    classes=['overlay']
class cta_node(BaseClassNode):
    classes=['cta']
class number_node(BaseClassNode):
    classes=['number']
class benefit_node(BaseClassNode):
    classes=['benefit']
class text_cols_div_node(BaseClassNode):
    classes=['text-cols']
class embed_node(BaseClassNode):
    classes=['embed']
class youtube_node(BaseClassNode):
    youtube_attributes = {
        'data-youtube': None,
        'data-autoplay': None,
        'data-no-controls': None,
    }

    def __init__(self, youtube_id: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.youtube_attributes['data-youtube-id'] = youtube_id

    def show_controls(self):
        del(self.youtube_attributes['data-no-controls'])

class process_step_node(BaseClassNode):
    classes=['process']

    def __init__(self, *args, step_number: int = None, **kwargs):
        super().__init__(*args, **kwargs)
        if step_number is not None:
            self.add_step_number(step_number)

    def add_step_number(self, step_number: int):
        self.add_class(f"step-{step_number}")

class ProcessStepDirective(GenericDirective):
    required_arguments = 1
    node_type = process_step_node

    def run(self):
        node = super().run()[0]
        node.add_step_number(self.arguments[0])
        return [node]

class YoutubeDirective(GenericDirective):
    required_arguments = 1
    node_type = embed_node
    option_spec = copy.deepcopy(GenericDirective.option_spec)
    option_spec.update({
        'show-controls': directives.unchanged,
    })

    def run(self):
        node = super().run()[0]
        yt_node = youtube_node(self.arguments[0])
        if 'show-controls' in self.options:
            yt_node.show_controls()
        yt_node.add_attributes(yt_node.youtube_attributes)
        node += yt_node
        return [node]

ALLOWED_SIDEBAR_CONFIGS = [
    'sm', 'ms', 'sms'
]
class GridDirective(GenericDirective):
    
    node_type = grid_node
    option_spec = copy.deepcopy(GenericDirective.option_spec)
    option_spec.update({
        'alignment': directives.unchanged,
        'sidebar-config': directives.unchanged,
    })

    def run(self):
        node = super().run()[0]
        a = 'alignment'
        if a in self.options:
            requested_alignment = self.options[a]
            if requested_alignment == 'vertical':
                node.add_class('vertical-align')
            else:
                print(f'ERROR: Unknown grid alignment: {requested_alignment}, ignoring')
        sc = 'sidebar-config'
        if sc in self.options:
            requested_config = self.options[sc].lower()
            if requested_config in ALLOWED_SIDEBAR_CONFIGS:
                node.add_class(requested_config)
        return [node]

div_map = {
    'div': div_node,
    'div-wrap': div_wrap_node,
    'column': column_node,
    'center': center_node,
    'flex-content': flex_content_node,
    'content-left': content_left_node,
    'content-center': content_center_node,
    'content-right': content_right_node,
    'overlay': overlay_node,
    'cta': cta_node,
    'number': number_node,
    'benefit': benefit_node,
    'text-cols-div': text_cols_div_node,
}

def process_step_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    node = process_step_node()
    node.add_step_number(int(text))
    return [node], []

def setup_div(app):
    for label in div_map:
        node_type = div_map[label]
        app.add_node(node_type)
        add_role_and_directive(app, node_type, label)
    app.add_node(grid_node)
    app.add_node(process_step_node)
    app.add_node(embed_node)
    app.add_node(youtube_node)
    app.add_directive('grid', GridDirective)
    app.add_directive('process-step', ProcessStepDirective)
    app.add_role('process-step', process_step_role)
    app.add_directive('youtube', YoutubeDirective)
    # Convenience
    app.add_role('ps', process_step_role)

class DivTranslator(HTMLTranslator):
    pass

add_visit_depart(DivTranslator, grid_node.__name__, TAG)
add_visit_depart(DivTranslator, process_step_node.__name__, TAG)
add_visit_depart(DivTranslator, embed_node.__name__, TAG)
add_visit_depart(DivTranslator, youtube_node.__name__, TAG)
for node_type in div_map.values():
    add_visit_depart(DivTranslator, node_type.__name__, TAG)




from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator

from .common import (
    process_classes,
    GenericDirective,
    add_attribute_to_tag,
    create_role,
    create_directive,
    generic_visit,
    generic_depart,
    get_node,
    get_generic_translator_class,
    add_role_and_directive,
    BaseNode,
    BaseClassNode,
    add_visit_depart,
)

TAG = 'p'

class paragraph_node(BaseNode): pass
class text_subtitle_node(BaseClassNode, paragraph_node):
    classes=['text-subtitle']
class text_symbols_node(BaseClassNode, paragraph_node):
    classes=['text-symbols']
class text_intro_node(BaseClassNode, paragraph_node):
    classes=['text-intro']
class text_center_node(BaseClassNode, paragraph_node):
    classes=['aligncenter']
class text_year_node(BaseClassNode, paragraph_node):
    classes=['year']
class text_title_node(BaseClassNode, paragraph_node):
    classes=['title']
class text_summary_node(BaseClassNode, paragraph_node):
    classes=['summary']
class text_context_node(BaseClassNode, paragraph_node):
    classes=['text-context']
class text_pull_right_node(BaseClassNode, paragraph_node):
    classes=['text-pull-right']
class paragraph_style_node(BaseClassNode): pass

paragraph_map = {
    'paragraph': paragraph_node,
    'text-subtitle': text_subtitle_node,
    'text-symbols': text_symbols_node,
    'text-intro': text_intro_node,
    'text-center': text_center_node,
    'year': text_year_node,
    'title': text_title_node,
    'summary': text_summary_node,
    'text-context': text_context_node,
    'text-pull-right': text_pull_right_node,
}

class ParagraphStyleDirective(GenericDirective):
    node_type = paragraph_style_node
    required_arguments = 1
    has_content = True

    def run(self):
        # Don't need the logic of super()
        node = paragraph_style_node()
        node['style'] = self.arguments[0]
        if self.has_content:
            node = self._process_content(node)
        return [node]

def setup_paragraph(app):
    for label in paragraph_map:
        node_type = paragraph_map[label]
        app.add_node(node_type)
        add_role_and_directive(app, node_type, label)
    app.add_directive('paragraph-style', ParagraphStyleDirective)
    # Convenience label
    add_role_and_directive(app, text_intro_node, 'ti')
    add_role_and_directive(app, text_center_node, 'c')
    add_role_and_directive(app, text_center_node, 'tsym')
    add_role_and_directive(app, text_subtitle_node, 'tsub')
    add_role_and_directive(app, text_context_node, 'tc')
    add_role_and_directive(app, text_pull_right_node, 'tpr')

def get_tag_and_heading(node_text: str):
    count = 1
    while node_text[count] == '#':
        count += 1
    return (f'h{count}', '#' * count)

class ParagraphTranslator(HTMLTranslator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.paragraph_style = None
    
    def visit_paragraph(self, node):
        """ Filter blank paragraphs

        webslides is very sensitive to blank paragraphs
        AND sphinx really likes to produce them.
        """
        if node.astext() == '':
            return
        else:
            super().visit_paragraph(node)
            if '<p' in self.body[-1]:
                if self.paragraph_style is not None:
                    self.body[-1] = add_attribute_to_tag(
                        self.body[-1], 
                        'p', 
                        'style', 
                        self.paragraph_style
                    )

    def visit_Text(self, node):
        fail = True
        if len(node.astext()) > 0:
            if node.astext()[0] == '#':
                markdown_tag, markdown_heading = get_tag_and_heading(node.astext())            
                self.body.append(f"<{markdown_tag}>")
                super().visit_Text(
                    nodes.Text(node.astext().replace(markdown_heading, ''))
                )
                fail = False
        if fail:
            super().visit_Text(node)

    def depart_Text(self, node):
        fail = True
        if len(node.astext()) > 0:
            if node.astext()[0] == '#':
                markdown_tag, markdown_heading = get_tag_and_heading(node.astext())            
                self.body.append(f"</{markdown_tag}>")
                super().depart_Text(
                    nodes.Text(node.astext().replace(markdown_heading, ''))
                )
                fail = False
        if fail:
            super().depart_Text(node)

    def depart_paragraph(self, node):
        if node.astext() == '':
            return
        else:
            super().depart_paragraph(node)

    def visit_paragraph_style_node(self, node):
        self.paragraph_style = node['style']

    def depart_paragraph_style_node(self, node):
        self.paragraph_style = None


for thing in paragraph_map.values():
    add_visit_depart(
        ParagraphTranslator,
        thing.__name__,
        TAG,
    )


from copy import deepcopy

from sphinx.util.docutils import SphinxDirective
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.writers.html import HTMLTranslator
from sphinx.util import logging

from .div import div_wrap_node, div_node
from .span import background_node
from .video import video_node
from .header_footer import make_header_node, set_alignment, make_footer_node
from .paragraph import paragraph_node
from .common import (
    add_class_to_tag,
    process_classes,
    GenericDirective,   
    BaseClassNode,
)

logger = logging.getLogger(__name__)

SECTION_TAG = 'section'

def source_read_handler(app, docname, source):
    app.config['slide_config'] = {}

def setup_slide(app):
    app.add_node(slide_node)
    app.add_directive('slide', Slide)
    app.add_directive('slide-config', SlideConfig)
    app.connect('source-read', source_read_handler)

class slide_node(BaseClassNode):
    pass

class SlideTranslator(HTMLTranslator):
    slide_open = False
    
    def visit_slide_node(self, node):
        if self.slide_open:
            self.body.append(f'</{SECTION_TAG}>')
        content = process_classes(f'<{SECTION_TAG}>', node, SECTION_TAG)
        self.body.append(content)

    def depart_slide_node(self, node):
            self.body.append('</section>')
            self.slide_open = False 

class Slide(GenericDirective):
    node_type = slide_node
    option_spec = deepcopy(GenericDirective.option_spec)
    option_spec.update({
        # wrap will be deprecated
        'wrap': directives.unchanged,
        'no-wrap': directives.unchanged,
        'wrap-size': directives.unchanged,
        'wrap-zoom-in': directives.unchanged,
        'background-image': directives.unchanged,
        'dark-background-image': directives.unchanged,
        'light-background-image': directives.unchanged,
        'background-image-location': directives.unchanged,
        'background-image-animation': directives.unchanged,
        'background-color': directives.unchanged,
        'vertical-alignment': directives.unchanged,
        'content-alignment': directives.unchanged,
        'card-size': directives.unchanged,
        'card-background': directives.unchanged,
        'flex-content': directives.unchanged,
        'full-screen': directives.unchanged,
        'background-video': directives.unchanged,
        'background-video-poster': directives.unchanged,
        'background-video-dark': directives.unchanged,
        'text-serif': directives.unchanged,
        'no-defaults': directives.unchanged,
        'no-header': directives.unchanged,
        'header': directives.unchanged,
        'header-alignment': directives.unchanged,
        'no-footer': directives.unchanged,
        'footer': directives.unchanged,
        'footer-alignment': directives.unchanged,
    })

    def run(self):        
        if 'no-defaults' not in self.options:
            # get defaults and update with specific requests for this directive
            # i.e. requests override defaults
            slide_settings = deepcopy(self.env.app.config['slide_config'])
            slide_settings.update(
                self.options
            )
            self.options = slide_settings
        node = slide_node()
        active_node = node
        # Slide Node modifcations        
        active_node = self.set_horizontal_alignment(active_node)
        active_node = self.set_vertical_alignment(active_node)
        active_node = self.set_class_options(active_node)
        active_node = self.set_background_color(active_node)
        active_node = self.set_full_screen(active_node)
        active_node = self.set_text_serif(active_node)
        # Adding child nodes
        # Add before wrap
        active_node = self.set_background_image(active_node)        
        active_node = self.set_background_video(active_node)
        active_node = self.set_header(active_node)
        active_node = self.set_footer(active_node)
        # wrap
        active_node = self.check_wrap(active_node)  
        # after wrap
        active_node = self.check_card(active_node)
        active_node = self._process_content(active_node)      
        return [node]

    def set_header(self, node):
        sh = 'slide_header'
        set_requested_header = True
        try:
            if 'no-header' not in self.options:
                node += deepcopy(self.env.app.config[sh])
                set_requested_header = False
        except AttributeError:
            pass
        if set_requested_header:
            self.set_header_footer('header', 'header-alignment', node, make_header_node)
        return node
    
    def set_footer(self, node):
        sf = 'slide_footer'
        set_requested_footer = True
        try:
            if 'no-footer' not in self.options:
                node += deepcopy(self.env.app.config[sf])
                set_requested_footer = False
        except AttributeError:
            pass
        if set_requested_footer:
            self.set_header_footer('footer', 'footer-alignment', node, make_footer_node)
        return node

    def set_header_footer(self, option_name, alignment_name, node, node_maker):
        if option_name in self.options:
            hf_node, hf_active_node = node_maker()
            # Need to wrap as paragraph when not using header
            # directive, paragraph wraps alignment span
            paragraph = paragraph_node()
            hf_active_node += paragraph
            hf_active_node = paragraph
            if alignment_name in self.options:
                hf_active_node = set_alignment(
                    {'horizontal-alignment': self.options[alignment_name]}, 
                    hf_active_node
                )
            hf_active_node += nodes.Text(self.options[option_name])
            node += hf_node
    
    def set_full_screen(self, node):
        fs = 'full-screen'
        if fs in self.options:
            node.add_class('fullscreen')
        return node
    
    def check_card(self, node):
        cs = 'card-size'
        active_node = node
        if cs in self.options:
            # TODO: probably need to verify size against
            # what is in the CSS
            requested_size = self.options[cs]
            div = div_node()
            node += div
            div.add_class(f'card-{requested_size}')
            active_node = div
            cb = 'card-background'
            if cb in self.options:
                div.add_class(self.options[cb])
        return active_node
    
    def set_background_color(self, node):
        bc = 'background-color'
        if bc in self.options:
            node.add_class(self.options[bc])
        return node

    def set_text_serif(self, node):
        ts = 'text-serif'
        if ts in self.options:
            node.add_class('text-serif')
        return node

    def set_background_video(self, node):
        bv = 'background-video'
        bvp = 'background-video-poster'
        if bv in self.options:
            video = video_node(
                poster_target=self.options.get(bvp, None),
                video_target=self.options[bv],
                dark=('background-video-dark' in self.options),
            )
            node += video
        return node

    def set_background_image(self, node):
        bi = 'background-image'
        if bi in self.options:
            background = background_node()
            background.set_background_image(
                self.options[bi]
            )
            node += background
            if 'dark-background-image' in self.options:
                background.set_dark()
            if 'light-background-image' in self.options:
                background.set_light()
            if 'background-image-animation' in self.options:
                background.set_animation()
            bil = 'background-image-location'
            if bil in self.options:
                background.add_location(self.options[bil])
        return node
    
    def check_wrap(self, node):
        active_node = node
        if 'no-wrap' not in self.options:
            div = div_wrap_node()
            active_node = div
            ws = 'wrap-size'
            if ws in self.options:
                div.add_class(f"size-{self.options[ws]}")
            node += div
            wzi = 'wrap-zoom-in'
            if wzi in self.options:
                div.add_class('zoomIn')
            if 'wrap' in self.options:
                logger.warning("\"wrap\" option on slide directive no longer used, will be removed soon!")
        ca = 'content-alignment'
        if ca in self.options:
            div = div_node()
            requested_alignment = self.options[ca]
            good_request = False
            if requested_alignment == 'left':
                div.add_class('content-left')
                good_request = True
            elif requested_alignment == 'center':
                div.add_class('content-center')
                good_request = True
            elif requested_alignment == 'right':
                div.add_class('content-right')
                good_request = True
            else:
                print(f'ERROR: Bad requested content alignment: {requested_alignment}, ignoring...')
            if good_request:
                active_node += div
                active_node = div
        return active_node

    def set_vertical_alignment(self, node):
        va = 'vertical-alignment'
        if va in self.options:
            requested_alignment = self.options[va].lower()
            if requested_alignment == 'top':
                node.add_class('slide-top')
            elif requested_alignment == 'bottom':
                node.add_class('slide-bottom')
            else:
                print(f'ERROR: Vertical alignment {requested_alignment} unknown, ignoring...')
        return node

class SlideConfig(Slide):

    def run(self):
        self.env.app.config['slide_config'] = deepcopy(self.options)
        return []